/*
** ACEpansion generic library header
**
** Copyright  2020 by Philippe Rimauro
**
** Compatible with:
**  - MorphOS/gcc
**  - AmigaOS/vbcc
*/

#include <clib/alib_protos.h>

#include <exec/resident.h>
#include <exec/libraries.h>
#ifdef __MORPHOS__
#include <exec/system.h>
#include <exec/interrupts.h>
#include <exec/rawfmt.h>
#include <libraries/query.h>
#endif

#include <proto/exec.h>
#include <proto/locale.h>

#include <acepansion/lib_header.h>

#include "acepansion.h"
#define CATCOMP_BLOCK
#include "generated/locale_strings.h" /* catcomp block */

struct LibBase
{
	struct Library          LibNode;
	APTR                    SegList;
	struct SignalSemaphore  BaseLock;
	BOOL                    InitFlag;
};

extern const char VersionTag[];

struct Library *SysBase;
struct Library *LocaleBase;
struct Locale  *Locale;
struct Catalog *Catalog;

#ifdef __MORPHOS__
struct Library *libCreate(struct Library *unused, APTR segList, struct Library *sysBase);
struct LibBase *libInit(struct LibBase *base, APTR segList, struct Library *SysBase);
struct Library *libOpen(void);
ULONG libClose(void);
APTR libFlush(void);
ULONG libReserved(void);

struct ACEpansionPlugin * LibCreatePlugin(struct LibBase *base, CONST_STRPTR *toolTypes, struct SignalSemaphore *sema);
VOID LibDeletePlugin(struct LibBase *base, struct ACEpansionPlugin *plugin);
BOOL LibActivatePlugin(struct LibBase *base, struct ACEpansionPlugin *plugin, BOOL status);
VOID LibReset(struct LibBase *base, struct ACEpansionPlugin *plugin);
VOID LibEmulate(struct LibBase *base, struct ACEpansionPlugin *plugin, ULONG *signals);
VOID LibWriteIO(struct LibBase *base, struct ACEpansionPlugin *plugin, USHORT port, UBYTE value);
VOID LibReadIO(struct LibBase *base, struct ACEpansionPlugin *plugin, USHORT port, UBYTE * value);
VOID LibGetAudio(struct LibBase *base, struct ACEpansionPlugin *plugin, SHORT * leftSample, SHORT *rightSample);
VOID LibPrinter(struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE data, BOOL strobe, BOOL *busy);
VOID LibCursor(struct LibBase *base, struct ACEpansionPlugin *plugin);
VOID LibAcknowledgeInterrupt(struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE *ivr);
#if API_VERSION >= 2
VOID LibJoystick(struct LibBase *base, struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData);
#if API_VERSION >= 3
VOID LibReturnInterrupt(struct LibBase *base, struct ACEpansionPlugin *plugin);
#if API_VERSION >= 4
VOID LibHostLightDeviceDiodePulse(struct LibBase *base, struct ACEpansionPlugin *plugin);
VOID LibHostLightDeviceButton(struct LibBase *base, struct ACEpansionPlugin *plugin, BOOL pressed);
#if API_VERSION >= 5
STRPTR * LibGetPrefsPlugin(struct LibBase *base, struct ACEpansionPlugin *plugin, APTR pool);
#if API_VERSION >= 6
BOOL LibWriteMem(struct LibBase *base, struct ACEpansionPlugin *plugin, USHORT address, UBYTE value);
VOID LibReadMem(struct LibBase *base, struct ACEpansionPlugin *plugin, USHORT address, UBYTE * value, BOOL opcodeFetch);
#if API_VERSION >=7
VOID LibHostMouseEvent(struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE buttons, SHORT deltaX, SHORT deltaY);
VOID LibHostGamepadList(struct LibBase *base, struct ACEpansionPlugin *plugin, STRPTR *list);
VOID LibHostGamepadEvent(struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE index, USHORT buttons, BYTE ns, BYTE ew, BYTE lx, BYTE ly, BYTE rx, BYTE ry);
VOID LibAnalogInput(struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE channel, UBYTE *value);
#endif
#endif
#endif
#endif
#endif
#endif
#else
__saveds struct Library *libCreate(__reg("a0") APTR segList, __reg("a6") struct Library *sysbase);
__saveds struct Library *libOpen(__reg("a6") struct LibBase *base);
__saveds ULONG libClose(__reg("a6") struct LibBase *base);
__saveds APTR libFlush(__reg("a6") struct LibBase *base);
__saveds ULONG libReserved(VOID);

__saveds struct ACEpansionPlugin * LibCreatePlugin(__reg("a6") struct LibBase *base, __reg("a1") CONST_STRPTR *toolTypes, __reg("a2") struct SignalSemaphore *sema);
__saveds VOID LibDeletePlugin(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin);
__saveds BOOL LibActivatePlugin(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") BOOL activate);
__saveds VOID LibReset(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin);
__saveds VOID LibEmulate(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") ULONG *signals);
__saveds VOID LibWriteIO(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") USHORT port, __reg("d1") UBYTE value);
__saveds VOID LibReadIO(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") USHORT port, __reg("a1") UBYTE *value);
__saveds VOID LibGetAudio(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") SHORT *leftSample, __reg("a2") SHORT *rightSample);
__saveds VOID LibPrinter(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") UBYTE data, __reg("d1") BOOL strobe, __reg("a1") BOOL *busy);
__saveds VOID LibCursor(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin);
__saveds VOID LibAcknowledgeInterrupt(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") UBYTE *ivr);
#if API_VERSION >= 2
__saveds VOID LibJoystick(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") BOOL com1, __reg("d1") BOOL com2, __reg("a1") UBYTE *ioData);
#if API_VERSION >= 3
__saveds VOID LibReturnInterrupt(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin);
#if API_VERSION >= 4
__saveds VOID LibHostLightDeviceDiodePulse(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin);
__saveds VOID LibHostLightDeviceButton(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") BOOL pressed);
#if API_VERSION >= 5
__saveds STRPTR * LibGetPrefsPlugin(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") APTR pool);
#if API_VERSION >= 6
__saveds BOOL LibWriteMem(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") USHORT address, __reg("d1") UBYTE value);
__saveds VOID LibReadMem(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") USHORT address, __reg("a1") UBYTE *value, __reg("d1") BOOL opcodeFetch);
#if API_VERSION >= 7
__saveds VOID LibHostMouseEvent(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") UBYTE buttons, __reg("d1") SHORT deltaX, __reg("d2") SHORT deltaY);
__saveds VOID LibHostGamepadList(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") STRPTR *list);
__saveds VOID LibHostGamepadEvent(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") UBYTE index, __reg("d1") USHORT buttons, __reg("d2") BYTE ns, __reg("d3") BYTE ew, __reg("d4") BYTE lx, __reg("d5") BYTE ly, __reg("d6") BYTE rx, __reg("d7") BYTE ry);
__saveds VOID LibAnalogInput(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") UBYTE channel, __reg("a0") UBYTE *value);
#endif
#endif
#endif
#endif
#endif
#endif
#endif

APTR expungeLib(struct LibBase *base);
BOOL allocLibResources(struct LibBase *base);
VOID freeLibResources(struct LibBase *base);

const char libName[] = LIBNAME;

#ifdef __MORPHOS__
const struct TagItem rTags[] =
{
	{ QUERYINFOATTR_NAME     , (IPTR)libName        },
	{ QUERYINFOATTR_IDSTRING , (IPTR)&VersionTag[1] },
	{ QUERYINFOATTR_COPYRIGHT, (IPTR)COPYRIGHT      },
	{ QUERYINFOATTR_DATE     , (IPTR)DATE           },
	{ QUERYINFOATTR_VERSION  , VERSION              },
	{ QUERYINFOATTR_REVISION , REVISION             },
	{ QUERYINFOATTR_SUBTYPE  , QUERYSUBTYPE_LIBRARY },
	{ TAG_END,  0 }
};

struct Resident romTag =
{
	RTC_MATCHWORD,
	&romTag,
	&romTag + 1,
	RTF_EXTENDED | RTF_PPC,
	VERSION,
	NT_LIBRARY,
	0,
	(char *)libName,
	(char *)&VersionTag[1],
	(APTR)libCreate,
	REVISION,
	(struct TagItem *)rTags
};
#else
struct Resident romTag =
{
	RTC_MATCHWORD,
	&romTag,
	&romTag + 1,
	0,
	VERSION,
	NT_LIBRARY,
	0,
	(char *)libName,
	(char *)&VersionTag[1],
	(APTR)libCreate
};
#endif

APTR jumpTable[] =
{
#ifdef __MORPHOS__
	(APTR)FUNCARRAY_BEGIN,
	(APTR)FUNCARRAY_32BIT_NATIVE,
#endif
	(APTR)libOpen,
	(APTR)libClose,
	(APTR)libFlush,
	(APTR)libReserved,
#ifdef __MORPHOS__
	(APTR)0xFFFFFFFF,
	(APTR)FUNCARRAY_32BIT_SYSTEMV,
#endif
	(APTR)LibCreatePlugin,
	(APTR)LibDeletePlugin,
	(APTR)LibActivatePlugin,
	(APTR)LibReset,
	(APTR)LibEmulate,
	(APTR)LibWriteIO,
	(APTR)LibReadIO,
	(APTR)LibGetAudio,
	(APTR)LibPrinter,
	(APTR)LibCursor,
	(APTR)LibAcknowledgeInterrupt,
#if API_VERSION >= 2
	(APTR)LibJoystick,
#if API_VERSION >= 3
	(APTR)LibReturnInterrupt,
#if API_VERSION >= 4
	(APTR)LibHostLightDeviceDiodePulse,
	(APTR)LibHostLightDeviceButton,
#if API_VERSION >= 5
	(APTR)LibGetPrefsPlugin,
#if API_VERSION >= 6
	(APTR)LibWriteMem,
	(APTR)LibReadMem,
#if API_VERSION >= 7
    (APTR)LibHostMouseEvent,
    (APTR)LibHostGamepadList,
    (APTR)LibHostGamepadEvent,
    (APTR)LibAnalogInput,
#endif
#endif
#endif
#endif
#endif
#endif
	(APTR)0xFFFFFFFF,
#ifdef __MORPHOS__
	(APTR)FUNCARRAY_END
#endif
};

#ifdef __MORPHOS__
struct TagItem libTags[] =
{
	{ LIBTAG_FUNCTIONINIT, (IPTR)jumpTable             },
	{ LIBTAG_LIBRARYINIT , (IPTR)libInit               },
	{ LIBTAG_MACHINE     , MACHINE_PPC                 },
	{ LIBTAG_BASESIZE    , sizeof(struct LibBase)      },
	{ LIBTAG_SEGLIST     , 0                           },
	{ LIBTAG_TYPE        , NT_LIBRARY                  },
	{ LIBTAG_NAME        , 0                           },
	{ LIBTAG_IDSTRING    , 0                           },
	{ LIBTAG_FLAGS       , LIBF_CHANGED | LIBF_SUMUSED },
	{ LIBTAG_VERSION     , VERSION                     },
	{ LIBTAG_REVISION    , REVISION                    },
	{ LIBTAG_PUBLIC      , TRUE                        },
	{ TAG_END            , 0                           },
};
#endif


#ifdef __MORPHOS__
struct LibBase * libInit(struct LibBase *base, APTR segList, UNUSED struct Library *sysbase)
{
	InitSemaphore(&base->BaseLock);
	base->SegList = segList;
	return base;
}


struct Library * libCreate(UNUSED struct Library *unused, APTR segList, struct Library *sysBase)
{
	SysBase = sysBase;

	libTags[4].ti_Data = (IPTR)segList;
	libTags[6].ti_Data = (IPTR)romTag.rt_Name;
	libTags[7].ti_Data = (IPTR)romTag.rt_IdString;

	return (NewCreateLibrary(libTags));
}
#else
__saveds struct Library* libCreate(__reg("a0") APTR segList, __reg("a6") struct Library *sysbase)
{
	struct LibBase *base = NULL;

	SysBase = sysbase;

	if ((base = (struct LibBase*)MakeLibrary(jumpTable, NULL, NULL, sizeof(struct LibBase), 0)))
	{
		base->LibNode.lib_Node.ln_Type = NT_LIBRARY;
		base->LibNode.lib_Node.ln_Name = romTag.rt_Name;
		base->LibNode.lib_Flags = LIBF_CHANGED | LIBF_SUMUSED;
		base->LibNode.lib_Version = VERSION;
		base->LibNode.lib_Revision = REVISION;
		base->LibNode.lib_IdString = romTag.rt_IdString;
		base->LibNode.lib_OpenCnt = 0;
		base->SegList = segList;
		InitSemaphore(&base->BaseLock);
		AddLibrary((struct Library*)base);
	}

	return (struct Library*)base;
}
#endif


APTR expungeLib(struct LibBase *base)
{
	APTR segList = NULL;

	ObtainSemaphore(&base->BaseLock);

	if(base->LibNode.lib_OpenCnt == 0)
	{
		freeLibResources(base);
		Forbid();
		Remove((struct Node *)base);
		Permit();
		segList = base->SegList;
		FreeMem((UBYTE*)base - base->LibNode.lib_NegSize, base->LibNode.lib_NegSize + base->LibNode.lib_PosSize);
		base = NULL;
	}
	else
    {
        base->LibNode.lib_Flags |= LIBF_DELEXP;
    }

	if(base)
    {
        ReleaseSemaphore(&base->BaseLock);
    }

	return segList;
}


BOOL allocLibResources(UNUSED struct LibBase *base)
{
    BOOL success = FALSE;

    if((LocaleBase = OpenLibrary("locale.library", 0L)))
    {
    	if((Locale = OpenLocale(NULL)))
        {
    		Catalog = OpenCatalog(Locale, LIBNAME".catalog",
                OC_Version, VERSION,
                TAG_DONE);

            success = TRUE;
    	}
    }

    if(success)
    {
        success = InitResources();
    }

    return success;
}


VOID freeLibResources(UNUSED struct LibBase *base)
{
	FreeResources();

    if(LocaleBase)
    {
        if(Locale)
        {
            if(Catalog)
            {
                CloseCatalog(Catalog);
                Catalog = NULL;
            }

            CloseLocale(Locale);
            Locale = NULL;
        }

        CloseLibrary(LocaleBase);
    }
}


#ifdef __MORPHOS__
struct Library * libOpen(VOID)
{
	struct LibBase *base = (struct LibBase *)REG_A6;
#else
__saveds struct Library * libOpen(__reg("a6") struct LibBase *base)
{
#endif
	struct Library *lib = (struct Library *)base;

	ObtainSemaphore(&base->BaseLock);

	if(!base->InitFlag)
	{
		if(allocLibResources(base))
        {
            base->InitFlag = TRUE;
        }
		else
		{
			freeLibResources(base);
			lib = NULL;
		}
	}

	if(lib)
	{
		base->LibNode.lib_Flags &= ~LIBF_DELEXP;
		base->LibNode.lib_OpenCnt++;
	}

	ReleaseSemaphore(&base->BaseLock);

	if(!lib)
    {
        expungeLib(base);
    }
	return lib;
}


#ifdef __MORPHOS__
ULONG libClose(VOID)
{
	struct LibBase *base = (struct LibBase *)REG_A6;
#else
__saveds ULONG libClose(__reg("a6") struct LibBase *base)
{
#endif
	ULONG ret = 0;

	ObtainSemaphore(&base->BaseLock);

	if(--base->LibNode.lib_OpenCnt == 0)
	{
		if(base->LibNode.lib_Flags & LIBF_DELEXP)
        {
            ret = (ULONG)expungeLib(base);
        }
	}

	if(ret == 0) ReleaseSemaphore(&base->BaseLock);
	return ret;
}


#ifdef __MORPHOS__
APTR libFlush(VOID)
{
	struct LibBase *base = (struct LibBase *)REG_A6;
#else
__saveds APTR libFlush(__reg("a6") struct LibBase *base)
{
#endif

	return(expungeLib(base));
}


#ifdef __MORPHOS__
ULONG libReserved(VOID)
#else
__saveds ULONG libReserved(VOID)
#endif
{
	return 0;
}


/*
** Library API
*/

#ifdef __MORPHOS__
struct ACEpansionPlugin * LibCreatePlugin(UNUSED struct LibBase *base, CONST_STRPTR *toolTypes, struct SignalSemaphore *sema)
#else
__saveds struct ACEpansionPlugin * LibCreatePlugin(__reg("a6") struct LibBase *base, __reg("a1") CONST_STRPTR *toolTypes, __reg("a2") struct SignalSemaphore *sema)
#endif
{
    return CreatePlugin(toolTypes, sema);
}

#ifdef __MORPHOS__
VOID LibDeletePlugin(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin)
#else
__saveds VOID LibDeletePlugin(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin)
#endif
{
    DeletePlugin(plugin);
}

#ifdef __MORPHOS__
BOOL LibActivatePlugin(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, BOOL activate)
#else
__saveds BOOL LibActivatePlugin(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") BOOL activate)
#endif
{
    return ActivatePlugin(plugin, activate);
}

#ifdef __MORPHOS__
VOID LibReset(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin)
#else
__saveds VOID LibReset(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin)
#endif
{
    Reset(plugin);
}

#ifdef __MORPHOS__
VOID LibEmulate(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, ULONG *signals)
#else
__saveds VOID LibEmulate(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") ULONG *signals)
#endif
{
    Emulate(plugin, signals);
}

#ifdef __MORPHOS__
VOID LibWriteIO(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, USHORT port, UBYTE value)
#else
__saveds VOID LibWriteIO(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") USHORT port, __reg("d1") UBYTE value)
#endif
{
    WriteIO(plugin, port, value);
}

#ifdef __MORPHOS__
VOID LibReadIO(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, USHORT port, UBYTE *value)
#else
__saveds VOID LibReadIO(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") USHORT port, __reg("a1") UBYTE *value)
#endif
{
    ReadIO(plugin, port, value);
}

#ifdef __MORPHOS__
VOID LibGetAudio(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, SHORT *leftSample, SHORT *rightSample)
#else
__saveds VOID LibGetAudio(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") SHORT *leftSample, __reg("a2") SHORT *rightSample)
#endif
{
    GetAudio(plugin, leftSample, rightSample);
}

#ifdef __MORPHOS__
VOID LibPrinter(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE data, BOOL strobe, BOOL *busy)
#else
__saveds VOID LibPrinter(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") UBYTE data, __reg("d1") BOOL strobe, __reg("a1") BOOL *busy)
#endif
{
    Printer(plugin, data, strobe, busy);
}

#ifdef __MORPHOS__
VOID LibCursor(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin)
#else
__saveds VOID LibCursor(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin)
#endif
{
    Cursor(plugin);
}

#ifdef __MORPHOS__
VOID LibAcknowledgeInterrupt(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE *ivr)
#else
__saveds VOID LibAcknowledgeInterrupt(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") UBYTE *ivr)
#endif
{
    AcknowledgeInterrupt(plugin, ivr);
}

#if API_VERSION >= 2

#ifdef __MORPHOS__
VOID LibJoystick(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData)
#else
__saveds VOID LibJoystick(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") BOOL com1, __reg("d1") BOOL com2, __reg("a1") UBYTE *ioData)
#endif
{
    Joystick(plugin, com1, com2, ioData);
}

#if API_VERSION >= 3

#ifdef __MORPHOS__
VOID LibReturnInterrupt(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin)
#else
__saveds VOID LibReturnInterrupt(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin)
#endif
{
    ReturnInterrupt(plugin);
}

#if API_VERSION >= 4

#ifdef __MORPHOS__
VOID LibHostLightDeviceDiodePulse(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin)
#else
__saveds VOID LibHostLightDeviceDiodePulse(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin)
#endif
{
#if API_VERSION < 7
    LightDeviceDiodePulse(plugin);
#else
    HostLightDeviceDiodePulse(plugin);
#endif
}

#ifdef __MORPHOS__
VOID LibHostLightDeviceButton(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, BOOL pressed)
#else
__saveds VOID LibHostLightDeviceButton(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") BOOL pressed)
#endif
{
#if API_VERSION < 7
    LightDeviceButton(plugin, pressed);
#else
    HostLightDeviceButton(plugin, pressed);
#endif
}

#if API_VERSION >= 5

#ifdef __MORPHOS__
STRPTR * LibGetPrefsPlugin(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, APTR pool)
#else
__saveds STRPTR * LibGetPrefsPlugin(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") APTR pool)
#endif
{
    return GetPrefsPlugin(plugin, pool);
}

#if API_VERSION >= 6

#ifdef __MORPHOS__
BOOL LibWriteMem(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, USHORT address, UBYTE value)
#else
__saveds BOOL LibWriteMem(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") USHORT address, __reg("d1") UBYTE value)
#endif
{
    return WriteMem(plugin, address, value);
}

#ifdef __MORPHOS__
VOID LibReadMem(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, USHORT address, UBYTE *value, BOOL opcodeFetch)
#else
__saveds VOID LibReadMem(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") USHORT address, __reg("a1") UBYTE *value, __reg("d1") BOOL opcodeFetch)
#endif
{
    ReadMem(plugin, address, value, opcodeFetch);
}

#if API_VERSION >= 7

#ifdef __MORPHOS__
VOID LibHostMouseEvent(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE buttons, SHORT deltaX, SHORT deltaY)
#else
__saveds VOID LibHostMouseEvent(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") UBYTE buttons, __reg("d1") SHORT deltaX, __reg("d2") SHORT deltaY)
#endif
{
    HostMouseEvent(plugin, buttons, deltaX, deltaY);
}

#ifdef __MORPHOS__
VOID LibHostGamepadList(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, STRPTR *list)
#else
__saveds VOID LibHostGamepadList(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("a1") STRPTR *list)
#endif
{
    HostGamepadList(plugin, list);
}

#ifdef __MORPHOS__
VOID LibHostGamepadEvent(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE index, USHORT buttons, BYTE ns, BYTE ew, BYTE lx, BYTE ly, BYTE rx, BYTE ry)
#else
__saveds VOID LibHostGamepadEvent(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") UBYTE index, , __reg("d1") USHORT buttons, __reg("d2") BYTE ns, __reg("d3") BYTE ew, __reg("d4") BYTE lx, __reg("d5") BYTE ly, __reg("d6") BYTE rx, __reg("d7") BYTE ry)
#endif
{
    HostGamepadEvent(plugin, index, buttons, ns, ew, lx, ly, rx, ry);
}

#ifdef __MORPHOS__
VOID LibAnalogInput(UNUSED struct LibBase *base, struct ACEpansionPlugin *plugin, UBYTE channel, UBYTE *value)
#else
__saveds VOID LibAnalogInput(__reg("a6") struct LibBase *base, __reg("a0") struct ACEpansionPlugin *plugin, __reg("d0") UBYTE channel, __reg("a0") UBYTE *value)
#endif
{
    AnalogInput(plugin, channel, value);
}

#endif
#endif
#endif
#endif
#endif
#endif


#ifndef __MORPHOS__
static __saveds VOID rawDoFmtCount(__reg("a3") ULONG *in) { (*in)++; }
#endif


static STRPTR makeToolTypeGeneric(APTR pool, CONST_STRPTR template, CONST_STRPTR key, IPTR val)
{
    STRPTR tooltype;
    ULONG count = 0;

#ifdef __MORPHOS__
    NewRawDoFmt(template, (APTR)RAWFMTFUNC_COUNT, (STRPTR)&count, key, val);
#else
    APTR args[2] = { (APTR)key, (APTR)val };

    RawDoFmt(template, args, (void (* CONST)(void))rawDoFmtCount, (STRPTR)&count);
#endif

    if((tooltype = AllocPooled(pool, count)))
    {
#ifdef __MORPHOS__
        NewRawDoFmt(template, RAWFMTFUNC_STRING, tooltype, key, val);
#else
        RawDoFmt(template, args, (void (* CONST)(void))"\x16\xc0Nu", tooltype);
#endif
    }

    return tooltype;
}


/*
** Get a string out of the catalog from its id
*/

STRPTR GetString(LONG id)
{
    LONG   *l;
    UWORD  *w;
    STRPTR  builtin;
    STRPTR  fromCatalog;

	l = (LONG *)CatCompBlock;

    while(*l != id )
    {
        w = (UWORD *)((ULONG)l + 4);
        l = (LONG *)((ULONG)l + (ULONG)*w + 6);
    }
    builtin = (STRPTR)((ULONG)l + 6);

    fromCatalog = GetCatalogStr(Catalog, id, builtin);

    if(*fromCatalog == '\0')
        return builtin;
    else
        return fromCatalog;
}


/*
** Allocate and format a string tooltype using the provided memory pool
*/

STRPTR MakeToolType(APTR pool, CONST_STRPTR key, CONST_STRPTR val)
{
    return makeToolTypeGeneric(pool, val ? "%s=%s" : "%s", key, (IPTR)val);
}


/*
** Allocate and format a numeric tooltype using the provided memory pool
*/

STRPTR MakeToolTypeNum(APTR pool, CONST_STRPTR key, LONG val)
{
    return makeToolTypeGeneric(pool, "%s=%ld", key, val);
}

